// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#include "configuration.h"
#include "ui_configuration.h"
#include "mainwindow.h"
#include "timingtweaks.h"

#include <QFileDialog>
#include <QSettings>
#include <QtDebug>

// This class is for global configuration settings

configuration::configuration(QWidget *parent) :	QWidget(parent), ui(new Ui::configuration)
	{
	ui->setupUi(this);
	connect(ui->pushButton_browseVNC, SIGNAL(clicked(bool)), this, SLOT(slot_browseVNCpath()));
	connect(ui->pushButton_browswHwList, SIGNAL(clicked(bool)), this, SLOT(slot_browseHwPath()));
	connect(ui->pushButton_exit, SIGNAL(clicked(bool)), this, SLOT(slot_exitClicked()));
	connect(ui->pushButton_saveSettings, SIGNAL(clicked(bool)), this, SLOT(slot_saveSettingsClicked()));
	connect(ui->checkBox_hidePWs, SIGNAL(stateChanged(int)), this, SLOT(slot_hidePasswordSettingChanged()));
	connect(ui->checkBox_defaultMSLogon, SIGNAL(stateChanged(int)), this, SLOT(slot_defaultMSLogonSettingChanged()));
	connect(ui->spinBox_ngraph, SIGNAL(valueChanged(int)), this, SLOT(slot_ngraphChanged(int)));
	connect(ui->spinBox_maxSession, SIGNAL(valueChanged(int)), this, SLOT(slot_maxSessionChanged(int)));
	connect(ui->volumeSlider, SIGNAL(sliderReleased()), this, SLOT(slot_volumeChanged()));
	connect(ui->pushButton_timingTweaks, SIGNAL(clicked(bool)), this, SLOT(slot_timingTweaksClicked()));
	connect(ui->comboBox_os, SIGNAL(currentTextChanged(QString)), this, SLOT(slot_defaultOSchanged(QString)));
	slot_hidePasswordSettingChanged();

	ui->comboBox_os->addItems(mwptr->osList);
	slot_defaultOSchanged(OS_DEFAULT);

	this->setStyleSheet("QWidget#configuration { background-color:"
								" qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,"
								" stop: 0 #CCE5FF, stop: 1 #EEFFFF); }");
	restoreSettings();
	this->hide();
	}

// Geometry is the only thing saved unless save settings is clicked
configuration::~configuration()
	{
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\Settings");
	mvc.setValue("ConfigGeometryPos", pos());
	mvc.setValue("ConfigGeometrySize", size());
	delete ui;
	}

void configuration::restoreSettings()
	{
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\Settings");
	if(mvc.contains("ConfigGeometryPos"))
		{
		QPoint pos = mvc.value("ConfigGeometryPos").toPoint();
		move(pos);
		}
	if(mvc.contains("ConfigGeometrySize"))
		{
		QSize size = mvc.value("ConfigGeometrySize").toSize();
		resize(size);
		}
	mwptr->setVolume(mvc.value("bingVolume", 100).toInt());
	ui->volumeSlider->setValue(static_cast<int>(mwptr->getVolume()));
	}

// -------------------------- public functions -------------------------


// Returns true on success
QString configuration::getFileName(QString title, QString defaultName, QStringList filters)
	{
	QString fn = QString();
// The various browse routines call this to get a file name.
	QFileDialog *fdialog = new QFileDialog(this);
	fdialog->setViewMode(QFileDialog::Detail);
	fdialog->setFileMode(QFileDialog::AnyFile);
	fdialog->setWindowTitle(title);
	fdialog->selectFile(defaultName);
	if(defaultName.isEmpty())
		fdialog->setDirectory("c:\\");
	else
		fdialog->setDirectory(QFileInfo(defaultName).absolutePath());
	fdialog->setNameFilters(filters);
	if(fdialog->exec())
		{
		fn = fdialog->selectedFiles()[0];
		}
	delete fdialog;
	return fn;
	}

// --------------------------- private slots ---------------------------

// Browse for path to list of names and hardware addresses to bue used for wake-on-lan
void configuration::slot_browseHwPath()
	{
	QStringList filters;
	filters << tr("txt files (*.txt)")
			  << tr("All files (*)");
	QString browse = getFileName("Hardware address file", mwptr->getHWlistPath(), filters);
	if(!browse.isEmpty())
		{
		browse = QDir::toNativeSeparators(browse);
		mwptr->setHWlistPath(browse);
		ui->lineEdit_hwList->setText(browse);
		}
	}

// Browse for path to vnc viewer
void configuration::slot_browseVNCpath()
	{
	QStringList filters;
	filters << tr("Exe files (*.exe)")
				<< tr("All files (*)");
	QString browse = getFileName("VNC path", mwptr->getVNCpath(), filters);
	if(!browse.isEmpty())
		{
		browse = QDir::toNativeSeparators(browse);
		mwptr->setVNCpath(browse);
		ui->lineEdit_VNCpath->setText(browse);
		}
	}

void configuration::slot_exitClicked()
	{
	close();
	}

void configuration::slot_hidePasswordSettingChanged()
	{
	bool hide = ui->checkBox_hidePWs->isChecked();
	mwptr->setHidePasswords(hide);
	ui->lineEdit_MSLogonPassword->setEchoMode(hide ? QLineEdit::Password : QLineEdit::PasswordEchoOnEdit);
	ui->lineEdit_VNCpassword->setEchoMode(hide ? QLineEdit::Password : QLineEdit::PasswordEchoOnEdit);
	ui->lineEdit_welcomePassword->setEchoMode(hide ? QLineEdit::Password : QLineEdit::PasswordEchoOnEdit);
	}
void configuration::slot_defaultMSLogonSettingChanged()
	{
	mwptr->defaultMSLogon = ui->checkBox_defaultMSLogon->isChecked();
	}

void configuration::slot_defaultOSchanged(QString newvalue)
	{
	int newnum = -1;

	for(int i=0; i<nOS; i++)
		{
		if(newvalue == mwptr->osList[i])
			{
			newnum = i;
			break;
			}
		}
	mwptr->setDefaultOS(newnum);
	ui->comboBox_os->setCurrentIndex(newnum);
	}

void configuration::slot_ngraphChanged(int newvalue) { mwptr->setNgraph(newvalue); }

void configuration::slot_maxSessionChanged(int newvalue) { mwptr->setMaxSession(newvalue); }

void configuration::slot_volumeChanged()
	{
	mwptr->setVolume(ui->volumeSlider->value());
	mwptr->playBing();
	}

void configuration::slot_timingTweaksClicked()
	{
	TimingTweaks *tweaks = new TimingTweaks(nullptr);
	tweaks->setAttribute(Qt::WA_DeleteOnClose);
	tweaks->show();
	tweaks->raise();
	mwptr->forceOnscreen(tweaks);
	}
