// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "pingresult.h"
#include "ui_pingresult.h"
#include "individualsettings.h"

//#include <iterator>
#include <QMessageBox>
#include <QProcess>
//#include <QQueue>
//#include <QSettings>
//#include <QSignalMapper>
//#include <QtDebug>
#include <QMenu>
#include <QThread>
#include <QTimeZone>

#include <WtsApi32.h>
// This class displays the graph of ping times
PingResult::PingResult(QWidget *parent, int taskno) : QWidget(parent), ui(new Ui::PingResult)
	{
	ui->setupUi(this);
	isetOpen = false;
	VNCpwType = VNC_DEFAULT;
	WOLsent = false;
	useRDP = false;
	deletionInProgress = false;
	signalMapperHost = nullptr;
	winsizeRDP = QPoint(0,0);
	this->setAttribute(Qt::WA_DeleteOnClose);
	this->setStyleSheet("QGroupBox { background-color:"
							  " qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,"
							  " stop: 0 #CCE5FF, stop: 1 #99FFFF); margin: 0 px; padding: 0 px; }"
							  "QPushButton { padding-left: 5 px; padding-right: 5 px; padding-top: 2 px; padding-bottom: 2 px; "
							  "margin: 0 px; border: 1 px solid black; border-style: outset; background-color: rgb(240,240,240); }");
	connect(ui->pushButton_WOL, SIGNAL(clicked(bool)), this, SLOT(slot_WOLclicked()));
	connect(ui->pushButton_VNCsmall, SIGNAL(clicked(bool)), this, SLOT(slot_VNCsmallClicked()));
	connect(ui->pushButton_VNClarge, SIGNAL(clicked(bool)), this, SLOT(slot_VNClargeClicked()));
	connect(ui->pushButton_settings, SIGNAL(clicked(bool)), this, SLOT(slot_settingsClicked()));
	connect(ui->pushButton_delete, SIGNAL(clicked(bool)), this, SLOT(slot_delClicked()));

	closeWindowActContext = new QAction(tr("Close window"), this);
	sendCredsContext = new QAction(tr("Send credentials"), this);
		sendCredsContext->setToolTip(tr("<qt>Use this for things like program installs that require admin rights. "
				"Make sure the cursor is in the right place so you don't type credentials into an application that doesn't need them.</qt>"));
	sendPwContext = new QAction(tr("Send password"), this);
		sendPwContext->setToolTip(tr("<qt>Use this when user name is remembered and only password is needed. "
				"Make sure the cursor is in the right place so you don't type a password into an application that doesn't need it.</qt>"));
	raiseActContext = new QAction(tr("Switch to"), this);
	loginActContext = new QAction(tr("Login"), this);
		loginActContext->setToolTip(tr("<qt>Login only makes sense for a VNC connection, not RDP</qt>"));
	logoffActContext = new QAction(tr("Logoff"), this);
	shutdownActContext = new QAction(tr("Shutdown"), this);
	lockActContext = new QAction(tr("Lock"), this);
	unlockActContext = new QAction(tr("Unlock"), this);
	queryLoggedOnActContext = new QAction(tr("Is anyone logged on?"), this);
		queryLoggedOnActContext->setToolTip(tr("This will fail for localhost"));
	queryUpTimeActContext = new QAction(tr("When was system booted?"), this);
		queryUpTimeActContext->setToolTip(tr("This will fail for localhost"));

	connect(closeWindowActContext, SIGNAL(triggered()), this, SLOT(slot_closeFromContextMenu()));
	connect(raiseActContext, SIGNAL(triggered()), this, SLOT(slot_raiseFromContextMenu()));
	connect(sendCredsContext, SIGNAL(triggered()), this, SLOT(slot_sendCredsFromContextMenu()));
	connect(sendPwContext, SIGNAL(triggered()), this, SLOT(slot_sendPwFromContextMenu()));
	connect(loginActContext, SIGNAL(triggered()), this, SLOT(slot_loginFromContextMenu()));
	connect(logoffActContext, SIGNAL(triggered()), this, SLOT(slot_logoffFromContextMenu()));
	connect(shutdownActContext, SIGNAL(triggered()), this, SLOT(slot_shutdownFromContextMenu()));
	connect(lockActContext, SIGNAL(triggered()), this, SLOT(slot_lockFromContextMenu()));
	connect(unlockActContext, SIGNAL(triggered()), this, SLOT(slot_unlockFromContextMenu()));
	connect(queryLoggedOnActContext, SIGNAL(triggered(bool)), this, SLOT(slot_LoggedOnUserFromContextMenu()));
	connect(queryUpTimeActContext, SIGNAL(triggered(bool)), this, SLOT(slot_upTimeFromContextMenu()));

	ui->plotWidget->clearGraphs();
	ui->plotWidget->addGraph();
	ui->plotWidget->graph(0)->setLineStyle(QCPGraph::lsNone);
	ui->plotWidget->graph(0)->setScatterStyle(QCPScatterStyle::ssDisc);
	ui->plotWidget->addGraph();
	ui->plotWidget->graph(1)->setLineStyle(QCPGraph::lsNone);
	ui->plotWidget->graph(1)->setScatterStyle(QCPScatterStyle::ssDiamond);
	ui->plotWidget->graph(1)->setPen(QPen(QColor(Qt::red)));
	ui->plotWidget->axisRect()->setAutoMargins(QCP::msNone);
	ui->plotWidget->axisRect()->setMargins(QMargins(30,1,1,1));
	ui->plotWidget->xAxis->setTickLabels(false);
	ui->plotWidget->yAxis->setAutoTickCount(2);
	plotData(false);

	ui->label_downTime->setStyleSheet("QLabel { color: red; }");

	setTaskNum(taskno);
	}

PingResult::~PingResult()
	{
	delete ui;
	}

// Considered down if last 2 pings were not answered
// Return invalid if no host entered or name server lookup failed.
// If no data, return up.
// If only 1 ping available, use that one.
targetState PingResult::isUp()
	{
	if(getHost().isEmpty())
		return STATE_INVALID;
	if("Lookup failed" == getIPtext())
		return STATE_INVALID;
	if(pingData.isEmpty())
		return STATE_UP;
	QLinkedList<unsigned long>::iterator j = pingData.end();
	if(NO_PING == *(j-1))
		{
		if(pingData.count() > 1)
			{
			if(NO_PING == *(j-2))
				return STATE_DOWN;
			}
		return STATE_DOWN;
		}
	WOLsent = false;
	return STATE_UP;
	}

// -------------------------- public functions -------------------------

// Take queued ping times and put them in the data array for plotting.
// When the array is filled, discard the oldest values to make room for the new.
void PingResult::processData(bool sound)
	{
	if(pingQueue.isEmpty())
		return;
	while(!pingQueue.isEmpty())
		{
		if(pingData.count() >= mwptr->getNgraph())
			pingData.removeFirst();
		pingData.append(pingQueue.dequeue());
		}
	plotData(sound);
	}

// Set the task number which is used as the primary identifier of the task
void PingResult::setTaskNum(int newnum)
	{
	mytask = newnum;
// If host is changed we need to close settings and tell the main window which task changed.
	disconnect(ui->lineEdit_host, SIGNAL(editingFinished()));
	connect(ui->lineEdit_host, SIGNAL(editingFinished()), this, SLOT(slot_closeSettings()));
	delete signalMapperHost;
	signalMapperHost = new QSignalMapper(this);
	connect(ui->lineEdit_host, SIGNAL(editingFinished()), signalMapperHost, SLOT(map()));
	signalMapperHost->setMapping(ui->lineEdit_host, mytask);
	connect(signalMapperHost, SIGNAL(mapped(int)), this, SIGNAL(newHostSignal(int)));
	}

// Color the Wake button.
//   Green for up
//   Red for down
//   Magenta for wake-on-lan sent
//   Qt default if no host name has been entered or not pinging
// Also disable the button if already awake or no target set
void PingResult::setButtonColor(enum buttonsColors color)
	{
	bool buttonEnabled = true;
	switch(color)
		{
		case DEFAULT_COLOR:
			buttonEnabled = false;
			ui->pushButton_WOL->setStyleSheet("");
			break;
		case UP_COLOR:
			buttonEnabled = false;
			ui->pushButton_WOL->setStyleSheet(
				"background: qlineargradient(x1: 0 y1: 0, x2: 0, y2: 1,"
				"stop: 0 #00f000, stop: 1 #00c000);");
			break;
		case WOL_COLOR:
			ui->pushButton_WOL->setStyleSheet(
				"background: qlineargradient(x1: 0 y1: 0, x2: 0, y2: 1,"
				"stop: 0 #e000e0, stop: 1 #c000c0);");
			break;
		case DOWN_COLOR:
			ui->pushButton_WOL->setEnabled(true);
			ui->pushButton_WOL->setStyleSheet(
				"background: qlineargradient(x1: 0 y1: 0, x2: 0, y2: 1,"
				"stop: 0 #f00000, stop: 1 #c00000);");
			break;
		}
	ui->pushButton_WOL->setEnabled(buttonEnabled);
	}

// Set button labels to indicate RDP or VNC
void PingResult::updateButtonLabels()
	{
	QString host = ui->lineEdit_host->text();
	useRDP = mwptr->pingResults[mytask]->useRDP;
	if(useRDP)
		{
		ui->pushButton_VNClarge->setText("RDP FS");
		ui->pushButton_VNCsmall->setText("RDP window");
		}
	else
		{
		ui->pushButton_VNClarge->setText("VNC Large");
		ui->pushButton_VNCsmall->setText("VNC Small");
		}
	}

void PingResult::vncClicked(bool tiny)
	{
	if(getHost().isEmpty())
		{
		QMessageBox::warning(this, tr("Target needed"), tr("Please enter a host name"));
		return;
		}
	QStringList VNCargs;
	if(tiny)
		{
		VNCargs << "/scale" << mwptr->getVNCScale();
		if(mwptr->getViewOnlySmall())
			VNCargs << "/viewonly";
		}
	else
		VNCargs << "/autoscaling";

	int pwType = VNCpwType;
	if(VNC_DEFAULT == pwType)
		pwType = mwptr->getDefaultMSLogon() ? VNC_MS_DOMAIN : VNC_COMMON;
	switch(pwType)
		{
		case VNC_COMMON:
			if(needVNCpw())
				return;
			VNCargs << "/password" << mwptr->getVNCpassword();
			break;
		case VNC_MS_DOMAIN:
				if(needMSpw())
					return;
			VNCargs << "/user" << mwptr->getMSLogonName() << "/password" << mwptr->getMSLogonPassword();
			break;
		case VNC_CUSTOM:
			if(needCustomVNCpassword())
				return;
			VNCargs << "/password" << customVNCpassword;
		}
	VNCargs << getHost();
	QProcess *VNCrun = new QProcess();
	VNCrun->startDetached(mwptr->getVNCpath(), VNCargs);
	delete VNCrun;
	}

// Make a VNC connection with a small or full-size window
void PingResult::slot_VNClargeClicked()
	{
	if(useRDP)
		rdpClicked(false);
	else
		vncClicked(false);
	}

void PingResult::slot_VNCsmallClicked()
	{
	if(useRDP)
		rdpClicked(true);
	else
		vncClicked(true);
	}

// -------------------- public getters and setters ---------------------

// Set the host name and change button labels to indicate RDP or VNC
void PingResult::setHost(QString host) { ui->lineEdit_host->setText(host); }

// Get or set the ip address in text format
QString PingResult::getIPtext() { return ui->label_IP->text(); }
void PingResult::setIPtext(QString iptext) { ui->label_IP->setText(iptext); }

// Clear all ping time history for this host
void PingResult::resetData() { pingData.clear(); }

// Return the host name
QString PingResult::getHost() { return ui->lineEdit_host->text(); }

// Record or clear that wake-on-lan has been sent (and not yet responded to)
void PingResult::setWOLsent() { WOLsent = true; }
void PingResult::clearWOLsent() { WOLsent = false; }

// Has WOL been sent?
bool PingResult::wakeSent() { return WOLsent; }

// Get or set custom VNC password for this task
QString PingResult::getCustomVNCpassword() { return customVNCpassword; }
void PingResult::setCustomVNCpassword(QString pw) { customVNCpassword = pw; }

// Get or set last up time
QDateTime PingResult::getLastUpTime() { return lastUpTime; }
void PingResult::setLastUpTime(QDateTime dt) { lastUpTime = dt; }

// Get or set useRDP (as opposed to using VNC)
bool PingResult::getRDPstatus() { return useRDP; }
void PingResult::setRDPstatus(bool newstatus) { useRDP = newstatus; }

int PingResult::getCustomOS() { return customOS; }
void PingResult::setCustomOS(int newos) { customOS = newos; }

bool PingResult::getSlowness() { return slowPC; }
void PingResult::setSlowness(bool slow) { slowPC = slow; }

// Get or set size for RDP window
QPoint PingResult::getRDPsize() { return winsizeRDP; }
void PingResult::setRDPsize(QPoint xy) { winsizeRDP = xy; }

// Record the type of credentials needed for this target for VNC
void PingResult::setVNCpwType(VNCpasswordType pwt) { VNCpwType = pwt; }
VNCpasswordType PingResult::getVNCpwType() { return VNCpwType; }

// Get or set use welcome screen status
bool PingResult::getWelcomeStatus() { return useWelcome; }
void PingResult::setWelcomeStatus(bool newstatus) { useWelcome = newstatus; }

// Get or set user position on welcome screen
void PingResult::setWelcomeUserNumber(int user) {welcomeUserNumber = user; }
int PingResult::getWelcomeUserNumber() {return welcomeUserNumber; }

QString PingResult::getCustomWindowsName() { return customWindowsName; }
QString PingResult::getCustomWindowsPW() {return customWindowsPW; }
void PingResult::setCustomWindowsName(QString username) { customWindowsName = username; }
void PingResult::setCustomWindowsPW(QString pw) { customWindowsPW = pw; }

warnSetting PingResult::getWarnSetting() { return localWarn; }
void PingResult::setWarnSetting(warnSetting warn) { localWarn = warn; }

// ------------------------- private functions -------------------------

void PingResult::rdpClicked(bool windowed)
	{
	QStringList RDPargs;
	RDPargs << "/v:" << getHost();
	if(windowed)
		RDPargs << "/w:" << QString::number(winsizeRDP.x()) << "/h:" << QString::number(winsizeRDP.y());
	else
		RDPargs << "/f";
	QProcess *VNCrun = new QProcess();
	VNCrun->startDetached(QString::fromLocal8Bit(qgetenv("windir")) + "/system32/mstsc.exe", RDPargs);
	delete VNCrun;
	}

// VNC was requested, check if needed global credentials have been entered
bool PingResult::needVNCpw()
	{
	if(mwptr->getVNCpassword().isEmpty())
		{
		QMessageBox::warning(this, tr("Credentials needed"), tr("Enter global VNC credentials and try again"));
		mwptr->slot_configClicked();
		return true;
		}
	return false;
	}

// VNC was requested, check if needed custom credentials have been entered
bool PingResult::needCustomVNCpassword()
	{
	if(customVNCpassword.isEmpty())
		{
		QMessageBox::warning(this, tr("Credentials needed"), tr("Enter custom VNC credentials and try again"));
		slot_settingsClicked();
		return true;
		}
	return false;
	}

// Check if needed MS logon credentials are needed
// Returns true if credentials were not entered
bool PingResult::needMSpw()
	{
	if(customWindowsName.isEmpty())
		{
		if(mwptr->getMSLogonName().isEmpty() || mwptr->getMSLogonPassword().isEmpty())
			{
			QMessageBox::warning(this, tr("Credentials needed"), tr("Enter MS login credentials and try again"));
			mwptr->slot_configClicked();
			return true;
			}
		return false;
		}
	if(customWindowsPW.isEmpty())
		{
		QMessageBox::warning(this, tr("Credentials needed"), tr("Enter custom login password and try again"));
		slot_settingsClicked();
		return true;
		}
	return false;
	}

// Plot ping times using qcustomplot
void PingResult::plotData(bool sound)
	{
	if(pingData.count() < 1)
		return;
	QVector<double> x;
	QVector<double> y;
	QVector<double> xbad;
	QVector<double> ybad;
	QLinkedList<unsigned long>::iterator j = pingData.begin();
	for(int i=0; i<pingData.count(); i++, j++)
		{
		if(INVALID_PING == *j)
			continue;
		if(NO_PING == *j)
			{
			xbad.append(i);
			ybad.append(0);
			}
		else
			{
			x.append(i);
			y.append(*j);
			}
		}

	ui->plotWidget->xAxis->setRange(0, mwptr->getNgraph());
	ui->plotWidget->graph(0)->setData(x, y);
	ui->plotWidget->graph(1)->setData(xbad, ybad);
	ui->plotWidget->yAxis->setRange(0, 1);
	ui->plotWidget->graph(0)->rescaleAxes(true);
	ui->plotWidget->graph(1)->rescaleAxes(true);
	ui->plotWidget->yAxis->scaleRange(1.1, ui->plotWidget->yAxis->range().center());
	switch(isUp())
		{
		case STATE_UP:
			ui->plotWidget->axisRect()->setBackground(QBrush(QColor(230, 255, 230)));
			break;
		case STATE_DOWN:
			ui->plotWidget->axisRect()->setBackground(QBrush(QColor(255, 230, 230)));
			break;
		case STATE_INVALID:
			break;
		}
	ui->plotWidget->replot();

// Play an alert (if enabled) when last 3 results (ignoring invalids) are
// down/down/up or up/down/down.  In other words, when ping state changes
// but ignore single missed pings.
	if(!sound)
		return;

	int stateCount = 0;
	targetState lastThree[3];
	for(QLinkedList<unsigned long>::iterator k=pingData.end(); k!=pingData.begin();)
		{
		k--;
		if(INVALID_PING == *k)
			continue;
		if(NO_PING == *k)
			lastThree[stateCount] = STATE_DOWN;
		else
			lastThree[stateCount] = STATE_UP;
		stateCount++;
		if(3 == stateCount)
			break;
		}
	if(stateCount < 3)
		return;
	if(STATE_UP == lastThree[1])
		return;
	if(lastThree[0] == lastThree[2])
		return;
	mwptr->playBing();
	}

// --------------------------- public slots ----------------------------

// We have a result.  Put it in a FIFO queue for processing
void PingResult::slot_newData(unsigned long pingTime)
	{
	pingQueue.enqueue(pingTime);
	}

// --------------------------- private slots ---------------------------

// This gets called if the host name is edited so there's no confusion
// as to which name the settings apply to.
void PingResult::slot_closeSettings()
	{
	if(isetOpen)
		{
		iset->close();
		isetOpen = false;
		}
	}

// When delete is clicked, record the fact that we are being deleted so leaning
// on the mouse button won't generate multiple deletes for this task.  Also,
// if the individual settings dialog is open, it gets deleted automatically and
// we briefly return to slot_settingsClicked() and need to test if it is safe
// to call updateButtonLabels().
void PingResult::slot_delClicked()
	{
	if(!deletionInProgress)
		{
		deletionInProgress = true;
		emit delHostSignal(mytask);
		}
	}

// Open the configuration dialog (if we have a host name and not already open)
void PingResult::slot_settingsClicked()
	{
	if(getHost().isEmpty())
		return;
	if(isetOpen)
		return;
	iset = new IndividualSettings(this);
	iset->slot_setTask(mytask);
	iset->setAttribute(Qt::WA_DeleteOnClose, true);
	iset->setModal(true);
	isetOpen = true;
	iset->exec();
	if(!deletionInProgress)
		updateButtonLabels();
	isetOpen = false;
	}

// Tell the main window that the WOL button has been clicked.
void PingResult::slot_WOLclicked() { emit signalWOLsent(mytask); }

void PingResult::contextMenuEvent(QContextMenuEvent *event)
	{
	QMenu menu(this);
	menu.setToolTipsVisible(true);
	menu.addAction(closeWindowActContext);
	menu.addAction(raiseActContext);
	menu.addAction(sendCredsContext);
	menu.addAction(sendPwContext);
	menu.addAction(loginActContext);
	menu.addAction(logoffActContext);
	menu.addAction(shutdownActContext);
	menu.addAction(lockActContext);
	menu.addAction(unlockActContext);
	menu.addAction(queryLoggedOnActContext);
	menu.addAction(queryUpTimeActContext);
// Login doesn't make sense because the RDP window closes when you log off
	if(useRDP)
		loginActContext->setDisabled(true);
	menu.exec(event->globalPos());
	}

void PingResult::slot_closeFromContextMenu()
	{
	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		SendMessage(targetWindow, WM_SYSCOMMAND, SC_CLOSE, 0);
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

void PingResult::slot_raiseFromContextMenu()
	{
	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		{
		ShowWindow(targetWindow, SW_RESTORE);
		SetForegroundWindow(targetWindow);
		int delayMult = getSlowness() ? 2 : 1;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->activateDelay));
		}
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

// These routines send current login credentials for use with things like
// program installations that need admin rights
// Send username and password
void PingResult::slot_sendCredsFromContextMenu()
	{
	sendCredsFromContextMenu(false);
	}

// Send only password
void PingResult::slot_sendPwFromContextMenu()
	{
	sendCredsFromContextMenu(true);
	}

// Send uername+password or just password
void PingResult::sendCredsFromContextMenu(bool pwOnly)
	{
	if(needMSpw())
		return;

	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		{
		SetForegroundWindow(targetWindow);
		int delayMult = getSlowness() ? 2 : 1;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->activateDelay));
		if(!pwOnly)
			{
			sendQstring(pickUsername());
			sendKey(VK_TAB, false, 0);
			QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->loginNameDelay));
			}
		sendQstring(pickPW());
		sendKey(VK_RETURN, false, 0);
		}
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

void PingResult::slot_loginFromContextMenu()
	{
	if(needMSpw())
		return;

	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		{
		SetForegroundWindow(targetWindow);
		int delayMult = getSlowness() ? 2 : 1;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->activateDelay));
		sendCAD(CAD_DELAY_MS);
		if(useWelcome)
			{
			sendKey(VK_LEFT, false, 0);
			sendKey(VK_LEFT, false, 0);
			sendKey(VK_LEFT, false, 0);
			sendKey(VK_LEFT, false, 0);
			int nuser = getWelcomeUserNumber() - 1;
			for(int i=0; i<nuser; i++)
				sendKey(VK_RIGHT, false, 0);
			QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->welcomePositionDelay));
			sendKey(VK_SPACE, false, 0);
			sendKey(VK_BACK, false, 0);
			QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->welcomeSelectDelay));
			sendQstring(pickPW());
			sendKey(VK_RETURN, false, 0);
			return;
			}
		sendKey(VK_RETURN, false, 0);
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->warningDismissDelay));
		sendQstring(pickUsername());
		sendKey(VK_TAB, false, 0);
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->loginNameDelay));
		sendQstring(pickPW());
		sendKey(VK_RETURN, false, 0);
		}
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

// Logoff using WTSLogoffSession if possible, otherwise fall back to stuffing keystrokes.
// Only Windows 7 Pro and Ultimate have the remote desktop server that is required to make
// this work.  (There is a hack to add RDP server to Windows Home.)
// WTSLogoffSession gets access denied on Windows 7 with default settings even though
// WTSEnumerateSessions works.  Logoff only works if
// DWORD HKLM\SYSTEM\CurrentControlSet\Control\Terminal Server\AllowRemoteRPC = 1
void PingResult::slot_logoffFromContextMenu()
	{
	const WCHAR *remote_host = reinterpret_cast<const WCHAR *>(getHost().utf16());
	const WCHAR *remote_folder = TEXT("SYSTEM\\CurrentControlSet\\Control\\Terminal Server");
// open the HKLM remote registry
	HKEY hkey_remote;
	LONG res = RegConnectRegistry(remote_host, HKEY_LOCAL_MACHINE, &hkey_remote);
	if(res != ERROR_SUCCESS)
		{
//		qDebug() << "Registry connect failed";
		return;
		}
	HKEY hkey_folder;
	res = RegOpenKeyEx(hkey_remote, remote_folder, 0, KEY_READ, &hkey_folder);
	if(res != ERROR_SUCCESS)
		{
//		qDebug() << "Registry open failed";
		return;
		}
	int AllowRemoteRPC;
	DWORD bufferSize = 4;
	res = RegGetValueA(hkey_folder, "", "AllowRemoteRPC", RRF_RT_DWORD, nullptr, &AllowRemoteRPC, &bufferSize);
	RegCloseKey(hkey_remote);
	RegCloseKey(hkey_folder);
	if(res != ERROR_SUCCESS)
		{
//		qDebug() << "Registry read failed";
		return;
		}
//	qDebug() << "AllowRemoteRPC =" << AllowRemoteRPC;
	if(1 == AllowRemoteRPC)
		{
// Ugly code to get rid of warnings about old-style casts and const-ness
		const unsigned short *hostUTF16 = getHost().utf16();
		unsigned short *hostCopy = new unsigned short [sizeof(hostUTF16)];
		memcpy_s(hostCopy, sizeof(hostUTF16), hostUTF16, sizeof(hostUTF16));
		HANDLE targetMachine = WTSOpenServer(reinterpret_cast<wchar_t *>(hostCopy));
		delete[] hostCopy;
		if(nullptr != targetMachine)
			{
			bool logoffResult = false;
			PWTS_SESSION_INFO ppSessionInfo = nullptr;
			DWORD pCount = 0;
			if(WTSEnumerateSessions(targetMachine, 0, 1, &ppSessionInfo, &pCount))
				{
//			qDebug() << "pCount =" << pCount;
				bool sessionFound = false;
				DWORD sessionID = 0;
				for(DWORD i=0; i<pCount; ++i)
					{
					WTS_SESSION_INFO si = ppSessionInfo[i];
					if (WTSActive == si.State)
						{
// If the current session is active – store its ID
						sessionID = si.SessionId;
						sessionFound = true;
						break;
						}
					}
// SessionFound will be false if no one was logged on
				if(sessionFound)
					{
					logoffResult = WTSLogoffSession(targetMachine, sessionID, false);
					}
//		else
//			qDebug() << "Session id not found";

				WTSFreeMemory(ppSessionInfo);
				WTSCloseServer(targetMachine);
				}
//		else
//			qDebug() << "WTSEnumerateSessions error " << GetLastError();

// If logoffResult is true, the logoff worked and we're done, otherwise
// fall through to alternate method.
			if(logoffResult)
				return;
			}
		}

// Fall-through to old code using keystroke stuffing
	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		{
		SetForegroundWindow(targetWindow);
		int delayMult = getSlowness() ? 2 : 1;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->activateDelay));
		if(getRDPstatus())
			{
			sendKey(VK_END, true, VK_ALT_PRESSED | VK_CONTROL_PRESSED);
			QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->logoffCadDelay));
			}
		else
			sendCAD(mwptr->logoffCadDelay);
// Key to send depends on OS
		int os = getCustomOS();
		if(os < 0)
			os = mwptr->getDefaultOS();
		char logoffkey;
		QString osn = mwptr->osList[os];
		if(osn == "Windows 7" || osn == "Windows 8")
			logoffkey = 'L';
		else if(osn == "Windows 8" || osn == "2012 Server")
			logoffkey = 'O';
// Windows 10, Windows 11, 2016 Server (not checked) or 2019 Server
		else
			logoffkey = 'S';
		sendKey(logoffkey, false, VK_ALT_PRESSED);
		}
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

void PingResult::slot_lockFromContextMenu()
	{
	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		{
		SetForegroundWindow(targetWindow);
		int delayMult = getSlowness() ? 2 : 1;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->activateDelay));
		if(getRDPstatus())
			{
			sendKey(VK_END, true, VK_ALT_PRESSED | VK_CONTROL_PRESSED);
			QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->lockingCadDelay));
			}
		else
			sendCAD(delayMult*mwptr->lockingCadDelay);
		sendKey('K', false, VK_ALT_PRESSED);
		}
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

// If target is already unlocked, this will probably lock it
void PingResult::slot_unlockFromContextMenu()
	{
	if(needMSpw())
		return;

	HWND targetWindow = findWindow(getHost(), nameIsIP);
	if(targetWindow)
		{
		SetForegroundWindow(targetWindow);
		int delayMult = getSlowness() ? 2 : 1;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->activateDelay));
		sendCAD(mwptr->unlockingCadDelay);
		sendQstring(pickPW());
		sendKey(VK_RETURN, false, 0);
		}
	else
		QMessageBox::warning(this, tr("Not found"), tr("%1 not found").arg(getHost()));
	}

void PingResult::slot_shutdownFromContextMenu()
	{
	if(STATE_UP != isUp())
		return;
	InitiateSystemShutdownEx(LPWSTR(getHost().utf16()), nullptr, 10, true, false, 0);
//	bool shutrc = InitiateSystemShutdownEx(LPWSTR(getHost().utf16()), nullptr, 10, true, false, 0);
//	if(!shutrc)
//		{
// 53 = bad network path
//		DWORD ecode = GetLastError();
//		QMessageBox::warning(this, "Shutdown failed", tr("Error code is %1 for %2").arg(ecode).arg(getHost()));
//		}
	}

void PingResult::slot_LoggedOnUserFromContextMenu()
	{
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
	logcheck(false);
	QApplication::restoreOverrideCursor();
	}

void PingResult::logcheck(bool multi)
	{
	if(getHost().isEmpty())
		{
		if(multi)
			return;
		QMessageBox::warning(this, tr("Target needed"), tr("Please enter a host name"));
		return;
		}
	if(STATE_UP != isUp())
		return;
	QString *result = new QString();
	if(!multi)
		{
		mwptr->userList.clear();
		mwptr->ui->statusBar->show();
		mwptr->ui->statusBar->showMessage("Querying " + getHost());
		}
	mwptr->isAnyoneLoggedIn(getHost(), pickUsername(), pickPW(), result);
	mwptr->userList.append(getHost() + ": " + *result);
	delete result;
	if(!multi)
		{
		mwptr->showUserList(VIEW_LOGGEDIN);
		mwptr->ui->statusBar->clearMessage();
		mwptr->ui->statusBar->hide();
		mwptr->adjustSize();
		}
	}

void PingResult::slot_upTimeFromContextMenu()
	{
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
	upcheck(false);
	QApplication::restoreOverrideCursor();
	}

void PingResult::upcheck(bool multi)
	{
	if(getHost().isEmpty())
		{
		if(multi)
			return;
		QMessageBox::warning(this, tr("Target needed"), tr("Please enter a host name"));
		return;
		}
	if(STATE_UP != isUp())
		return;
	QString *result = new QString();
	if(!multi)
		{
		mwptr->userList.clear();
		mwptr->ui->statusBar->show();
		mwptr->ui->statusBar->showMessage("Querying " + getHost());
		}
	mwptr->upTime(getHost(), pickUsername(), pickPW(), result);
// If the result contains any letters, it's some sort of error message
	if((*result).contains(QRegExp("[a-zA-Z]")))
		{
		mwptr->userList.append(getHost() + ": " + *result);
		}
	else
		{
// Date format from wmic query is yyyyMMddHHmmss.sssssszzzz where zzzz is minutes, negative for US
// Fractional seconds are ignored here
		int year = (*result).mid(0,4).toInt();
		int month = (*result).mid(4,2).toInt();
		int day = (*result).mid(6,2).toInt();
		int hour = (*result).mid(8,2).toInt();
		int minute = (*result).mid(10,2).toInt();
		int second = (*result).mid(12,2).toInt();
		int zoneminutes = (*result).mid(21,4).toInt();
		QDateTime bootTime = QDateTime(QDate(year, month, day), QTime(hour, minute, second), QTimeZone(zoneminutes*60)).toLocalTime();
		qint64 secondsUp = bootTime.secsTo(QDateTime::currentDateTime());
		qint64 daysUp = secondsUp/86400;
		secondsUp -= daysUp*86400;
		qint64 hoursUp = secondsUp/3600;
		secondsUp -= hoursUp*3600;
		qint64 minutesUp = secondsUp/60;
		mwptr->userList.append(QString("%1: uptime %2 days, %3 hours, %4 minutes, booted %5")
			.arg(getHost()).arg(daysUp).arg(hoursUp).arg(minutesUp).arg(bootTime.toString()));
		}
	delete result;
	if(!multi)
		{
		mwptr->showUserList(VIEW_UPTIME);
		mwptr->ui->statusBar->clearMessage();
		mwptr->ui->statusBar->hide();
		mwptr->adjustSize();
		}
	}

// Return the login username apropriate for the current host
// This shouldn't be called if using the welcome screen
QString PingResult::pickUsername()
	{
	if(customWindowsPW.isEmpty() || customWindowsName.isEmpty())
		return mwptr->getMSLogonName();

	return customWindowsName;
	}

// Return the login password apropriate for the current host
QString PingResult::pickPW()
	{
	if(customWindowsPW.isEmpty())
		return mwptr->getMSLogonPassword();

	if(useWelcome)
		return customWindowsPW;

	if(!customWindowsName.isEmpty())
		return customWindowsPW;

	return mwptr->getMSLogonPassword();
	}
