// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "pingtask.h"
#include "pingresult.h"

#include <QMessageBox>
#include <QTimer>
#include <QtDebug>
#include <QTime>
#include <QRandomGenerator>

#include <string.h>

#define DEFER(x) x
#ifdef _WIN32
	#define ICMP_ECHO_REPLY_3264 DEFER(ICMP_ECHO_REPLY)
	#define PICMP_ECHO_REPLY_3264 DEFER(PICMP_ECHO_REPLY)
#else
	#define ICMP_ECHO_REPLY_3264 DEFER(ICMP_ECHO_REPLY32)
	#define PICMP_ECHO_REPLY_3264 DEFER(PICMP_ECHO_REPLY32)
#endif


#define NPING_AT_ONCE	1

typedef VOID (WINAPI *PIO_APC_ROUTINE)
	(
	IN PVOID ApcContext,
	IN PIO_STATUS_BLOCK IoStatusBlock,
	IN ULONG Reserved
	);
void CALLBACK gotPing(PVOID buf, PIO_STATUS_BLOCK IoStatusBlock, ULONG Reserved);

// This class  does the pinging
PingTask::PingTask(QObject *parent, int taskno)
	{
	mytask = taskno;
	validIP = false;
	pingInProgress = false;
	hIcmpFile = nullptr;
	strncpy_s(SendData, 12, "Data Buffer", 12);  // 11 bytes plus terminating 0
	SendSize = 12;
	ReplySize = static_cast<DWORD>(NPING_AT_ONCE * (sizeof(ICMP_ECHO_REPLY_3264) + static_cast<size_t>(SendSize) + sizeof(IO_STATUS_BLOCK) + 8));
	ReplyApcContext = &mytask;
	connect(this, SIGNAL(signal_pingTime(int,ulong)), static_cast<MainWindow *>(parent), SLOT(slot_processPings(int,ulong)));
	timer = new QTimer();
	timer->setSingleShot(true);
	connect(timer, SIGNAL(timeout()), this, SLOT(slot_pingtest()), Qt::QueuedConnection);
	}

PingTask::~PingTask()
	{
	}

// -------------------------- public functions -------------------------

// Random time offset so all pings don't run at the same time
//void PingTask::startPing() { timer->start(qrand()%500); }
void PingTask::startPing() { timer->start(QRandomGenerator::global()->bounded(500)); }

// -------------------- public getters and setters ---------------------

// Set the task number which is used as the primary identifier of the task
void PingTask::setTaskNum(int newnum) { mytask = newnum; }

// Set or get the IP address being pinged
void PingTask::setIP(unsigned long ip) { ipaddr = ip; }
unsigned long PingTask::getIP() { return ipaddr; }

// Set or get the validity of the name and IP address.
// Nameserver lookup fail would be the likely reason for setting not valid.
void PingTask::setIPvalidity(bool valid) { validIP = valid; }
bool PingTask::getIPvalidity() { return validIP; }

// Set how often to ping and how long to wait for a response.
void PingTask::setPingTiming(int interval, int wait)
	{
	pingInterval = interval;
	pingWait = wait;
	}

// Need to know if we're waiting for a ping.  If so, deleting this
// instance or the reply buffer is probably a bad idea.
// Callback tells us when ping is completed.
void PingTask::setPingCompleted() { pingInProgress = false; }
bool PingTask::isPingInProgress() { return pingInProgress; }

// --------------------------- private slots ---------------------------

// Send out the ping requests.  They are returned asynchronously by a callback to gotPing.
void PingTask::slot_pingtest()
	{
	if(globalEnabled & validIP)
		{
// Opening a file dialog or moving the main window seems to delay the call to gotPing.
// (Or more likely, gotPing calls things in the main event loop which is blocked.)
// So don't send a ping until we've gotten a callback from the last one, and hope
// that gotPing always gets called eventually.
		if(pingInProgress)
			{
//			qDebug() << "pingInProgress already true when slot_pingtest called";
			}
		else
			{
			pingInProgress = true;
			ReplyBuffer = static_cast<LPVOID>(malloc(ReplySize));
			hIcmpFile = IcmpCreateFile();
			if (hIcmpFile == INVALID_HANDLE_VALUE)
				{
				QMessageBox::critical(nullptr, tr("Error"), tr("Unable to open handle.\n"
								  "IcmpCreatefile returned error: %1").arg(GetLastError()));
				return;
				}
			*static_cast<int *>(ReplyApcContext) = mytask;
//			qDebug() << "Sending ping, mytask =" << mytask;
			unsigned long icmpRc = IcmpSendEcho2(hIcmpFile, nullptr, static_cast<PIO_APC_ROUTINE>(gotPing), ReplyApcContext, ipaddr,
				SendData, static_cast<WORD>(SendSize), nullptr, ReplyBuffer, ReplySize, static_cast<WORD>(pingWait));
			if(0 != icmpRc)
				{
//				qDebug() << "Error code from IcmpSendEcho2 is" << icmpRc;
				}
// Expected return code is 0, then getLastError returns ERROR_IO_PENDING (997)
//			int err = GetLastError();
//			qDebug() << "Return code from IcmpSendEcho2 =" << rc << "error =" << err;
			}
		}
	else if(globalEnabled)
		emit signal_pingTime(mytask, INVALID_PING);

	if(globalEnabled)
		timer->start(pingInterval);
	}

// ------------------------------ callback -----------------------------

// This is called by Windows when a ping succeeds or times out.
// Timeout might come through as a response of unreachable from localhost
// or a timeout error.
void CALLBACK gotPing(PVOID buf, PIO_STATUS_BLOCK /* IoStatusBlock */, ULONG /* Reserved */)
	{
	int task = *static_cast<int *>(buf);
	PICMP_ECHO_REPLY_3264 pEchoReply = static_cast<PICMP_ECHO_REPLY_3264>(mwptr->pingTasks[task]->ReplyBuffer);
	DWORD nping = IcmpParseReplies(pEchoReply, sizeof(ICMP_ECHO_REPLY_3264) + sizeof(IO_STATUS_BLOCK) + 8);
	if(0 == nping)
		{
		mwptr->slot_processPings(task, NO_PING);
		goto CloseOut;
		}

	switch(pEchoReply->Status)
		{
		case IP_SUCCESS:
			if(nping > 0)
				{
				if(mwptr->pingTasks[task]->getIP() == pEchoReply->Address)
					{
					mwptr->slot_processPings(task, pEchoReply->RoundTripTime);
					}
				else
					{
					QMessageBox::warning(nullptr, "Reply from wrong IP",
						QString("Got response from %1\nexpected from %2")
						.arg(IPtoQString(pEchoReply->Address)).arg(IPtoQString(mwptr->pingTasks[task]->getIP())));
					}
				}
			else
				{
				mwptr->slot_processPings(task, NO_PING);
				QMessageBox::warning(nullptr, "Got IP_SUCCESS but nping = 0",
					QString("Ping of %1 was supposedly successful, but no ping was received")
					.arg(IPtoQString(pEchoReply->Address)));
				}
				break;

		case IP_DEST_NET_UNREACHABLE:		// 11002
		case IP_DEST_HOST_UNREACHABLE:	// 11003
		case IP_DEST_PROT_UNREACHABLE:	// 11004
		case IP_DEST_PORT_UNREACHABLE:	// 11005
		case IP_REQ_TIMED_OUT:				// 11010
		case IP_TTL_EXPIRED_TRANSIT:		// 11013
		case IP_TTL_EXPIRED_REASSEM:		// 11014
			mwptr->slot_processPings(task, NO_PING);
			break;

		case IP_BUF_TOO_SMALL:		// 11001
		case IP_NO_RESOURCES:		// 11006
		case IP_BAD_OPTION:			// 11007
		case IP_HW_ERROR:				// 11008
		case IP_PACKET_TOO_BIG:		// 11009
		case IP_BAD_REQ:				// 11011
		case IP_BAD_ROUTE:			// 11012
		case IP_PARAM_PROBLEM:		// 11015
		case IP_SOURCE_QUENCH:		// 11016
		case IP_OPTION_TOO_BIG:		// 11017
		case IP_BAD_DESTINATION:	// 11018
		case IP_GENERAL_FAILURE:	// 11050
		default:
			QMessageBox::warning(nullptr, "Unexpected error",
				QString("Got unexpected error %1 while pinging %2")
				.arg(pEchoReply->Status).arg(mwptr->pingTasks[task]->getIP()));
			break;
		}

CloseOut:
	mwptr->pingTasks[task]->cleanupPingDone();
//	qDebug() << QTime::currentTime() << "task" << task << "nping" << nping;
//	qDebug() << "gotPing called, task =" << task << "status =" << pEchoReply->Status << "nping =" << nping;
//	qDebug() << "IP address =" << QString("%1").arg(pEchoReply->Address,0,16) << "delay =" << pEchoReply->RoundTripTime;
	return;
	}

void PingTask::cleanupPingDone()
	{
	if(!hIcmpFile)
		{
		IcmpCloseHandle(hIcmpFile);
		hIcmpFile = nullptr;
		}
	free(ReplyBuffer);
	ReplyBuffer = nullptr;
	setPingCompleted();
	}
