// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

// All save/restore settings stuff goes here except for windows geometry.
// This is to make it easier to find where to make changes when a new
// option is added.

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "configuration.h"
#include "ui_configuration.h"
#include "individualsettings.h"
#include "ui_individualsettings.h"
#include "pingresult.h"

#include <QSettings>

//********************************* global configuration *************************
// Apply settings when anything is changed (even if not saved as future default)
void configuration::populateForm()
	{
	ui->lineEdit_VNCpath->setText(mwptr->getVNCpath());
	ui->lineEdit_hwList->setText(mwptr->getHWlistPath());
	ui->lineEdit_pingCycleTime->setText(QString::number(mwptr->getPingCycleTime()));
	ui->lineEdit_pingWaitTime->setText(QString::number(mwptr->getPingWaitTime()));
	ui->lineEdit_scale->setText(mwptr->getVNCScale());
	ui->lineEdit_username->setText(mwptr->getMSLogonName());
	ui->lineEdit_MSLogonPassword->setText(mwptr->getMSLogonPassword());
	ui->lineEdit_VNCpassword->setText(mwptr->getVNCpassword());
	ui->lineEdit_welcomePassword->setText(mwptr->getWelcomePassword());
	ui->spinBox_maxRows->setValue(mwptr->getMaxRows());
	ui->spinBox_maxCols->setValue(mwptr->getMaxCols());
	ui->spinBox_ngraph->setValue(mwptr->getNgraph());
	ui->spinBox_maxSession->setValue(mwptr->getMaxSession());
	ui->checkBox_smallViewOnly->setChecked(mwptr->getViewOnlySmall());
	ui->checkBox_defaultMSLogon->setChecked(mwptr->getDefaultMSLogon());
	ui->checkBox_allowDupes->setChecked(mwptr->getAllowDuplicates());
	ui->checkBox_hidePWs->setChecked(mwptr->getHidePasswords());
	ui->comboBox_os->setCurrentIndex(mwptr->getDefaultOS());
	ui->checkBox_warningPresent->setChecked(WARN_YES == mwptr->globalWarn);

	connect(ui->lineEdit_hwList, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_MSLogonPassword, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_pingCycleTime, SIGNAL(editingFinished()), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_pingWaitTime, SIGNAL(editingFinished()), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_scale, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_username, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_VNCpassword, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_welcomePassword, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->lineEdit_VNCpath, SIGNAL(textChanged(QString)), this, SLOT(slot_applySettings()));
	connect(ui->spinBox_maxRows, SIGNAL(valueChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->spinBox_maxCols, SIGNAL(valueChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->spinBox_ngraph, SIGNAL(valueChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->spinBox_maxSession, SIGNAL(valueChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->checkBox_warningPresent, SIGNAL(stateChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->checkBox_smallViewOnly, SIGNAL(stateChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->checkBox_defaultMSLogon, SIGNAL(stateChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->checkBox_allowDupes, SIGNAL(stateChanged(int)), this, SLOT(slot_applySettings()));
	connect(ui->comboBox_os, SIGNAL(currentIndexChanged(int)), this, SLOT(slot_applySettings()));
	}

// Apply settings to main window variables after enforcing
// restrictions on ping cycle and wait times.
void configuration::slot_applySettings()
	{
	int pwt = ui->lineEdit_pingWaitTime->text().toInt();
	int pct = ui->lineEdit_pingCycleTime->text().toInt();
	if(pwt < 1000)
		{
		pwt = 1000;
		ui->lineEdit_pingWaitTime->setText("1000");
		}
	if(100 + pwt > pct)
		{
		pct = 100 + pwt;
		ui->lineEdit_pingCycleTime->setText(QString::number(pct));
		}
	mwptr->setVNCpath(ui->lineEdit_VNCpath->text());
	mwptr->setHWlistPath(ui->lineEdit_hwList->text());
	mwptr->setPingCycleTime(pct);
	mwptr->setPingWaitTime(pwt);
	mwptr->setVNCScale(ui->lineEdit_scale->text());
	mwptr->setMSLogonName(ui->lineEdit_username->text());
	mwptr->setMSLogonPassword(ui->lineEdit_MSLogonPassword->text());
	mwptr->setVNCpassword(ui->lineEdit_VNCpassword->text());
	mwptr->setWelcomePassword(ui->lineEdit_welcomePassword->text());
	mwptr->setMaxRows(ui->spinBox_maxRows->value());
	mwptr->setMaxCols(ui->spinBox_maxCols->value());
	mwptr->setNgraph(ui->spinBox_ngraph->value());
	mwptr->setMaxSession(ui->spinBox_maxSession->value());
	mwptr->setViewOnlySmall(ui->checkBox_smallViewOnly->isChecked());
	mwptr->setDefaultMSLogon(ui->checkBox_defaultMSLogon->isChecked());
	mwptr->setAllowDuplicates(ui->checkBox_allowDupes->isChecked());
	mwptr->setHidePasswords(ui->checkBox_hidePWs->isChecked());
	mwptr->setDefaultOS(ui->comboBox_os->currentIndex());
	mwptr->globalWarn = ui->checkBox_warningPresent->isChecked() ? WARN_YES : WARN_NO;
	}

// Saved passwords are encrypted
void configuration::slot_saveSettingsClicked()
	{
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\Settings");
	mvc.setValue("ConfigGeometryPos", pos());
	mvc.setValue("ConfigGeometrySize", size());
	mvc.setValue("VNCpath", mwptr->getVNCpath());
	mvc.setValue("HWlistPath", mwptr->getHWlistPath());
	mvc.setValue("MSLogonName", mwptr->getMSLogonName());
	mvc.setValue("MSlogonPassword", encryptString(mwptr->getMSLogonPassword()));
	mvc.setValue("WelcomePassword", encryptString(mwptr->getWelcomePassword()));
	mvc.setValue("VNCpassword", encryptString(mwptr->getVNCpassword()));
	mvc.setValue("defaultMSLogon", mwptr->getDefaultMSLogon());
	mvc.setValue("VNCScale", mwptr->getVNCScale());
	mvc.setValue("pingCycleTime", mwptr->getPingCycleTime());
	mvc.setValue("maxRows", mwptr->getMaxRows());
	mvc.setValue("maxCols", mwptr->getMaxCols());
	mvc.setValue("Ngraph", mwptr->getNgraph());
	mvc.setValue("maxSession", mwptr->getMaxSession());
	mvc.setValue("viewOnlySmall", mwptr->getViewOnlySmall());
	mvc.setValue("allowDuplicates", mwptr->getAllowDuplicates());
	mvc.setValue("defaultOS", mwptr->osList[mwptr->getDefaultOS()]);
	mvc.setValue("globalWarn", static_cast<int>(mwptr->globalWarn));
	}

//***************************** individual machine configuration *********************
// This routine takes the values on the individual machine config and
// sends them to the appropriate pingResults task
void IndividualSettings::updatePwOption()
	{
// Remote connection
	mwptr->pingResults[mytask]->setRDPstatus(ui->radioButton_useRDP->isChecked());
	mwptr->pingResults[mytask]->setCustomOS(ui->comboBox_os->currentIndex() - 1);
	mwptr->pingResults[mytask]->setSlowness(ui->checkBox_slowPC->isChecked());

// VNC credentials
	if(ui->radioButton_VNCdefaultCreds->isChecked())
		mwptr->pingResults[mytask]->setVNCpwType(VNC_DEFAULT);
	else if(ui->radioButton_VNCpw->isChecked())
		mwptr->pingResults[mytask]->setVNCpwType(VNC_COMMON);
	else if(ui->radioButton_VNCmsLogin->isChecked())
		mwptr->pingResults[mytask]->setVNCpwType(VNC_MS_DOMAIN);
	else
		{
		mwptr->pingResults[mytask]->setVNCpwType(VNC_CUSTOM);
		mwptr->pingResults[mytask]->setCustomVNCpassword(ui->lineEdit_customVNCpassword->text());
		}

// Login warning present
	if(ui->radioButton_warningDefault->isChecked())
		mwptr->pingResults[mytask]->setWarnSetting(WARN_DEFAULT);
	else if(ui->radioButton_warningYes->isChecked())
		mwptr->pingResults[mytask]->setWarnSetting(WARN_YES);
	else
		mwptr->pingResults[mytask]->setWarnSetting(WARN_NO);

// Remote desktop options
	mwptr->pingResults[mytask]->
		setRDPsize(QPoint(ui->lineEdit_widthRDP->text().toInt(), ui->lineEdit_heightRDP->text().toInt()));

// Windows login
	mwptr->pingResults[mytask]->setWelcomeStatus(ui->checkBox_useWelcome->isChecked());
	mwptr->pingResults[mytask]->setWelcomeUserNumber(ui->spinBox_welcomePosition->value());
	mwptr->pingResults[mytask]->setCustomWindowsName(ui->lineEdit_customWindowsName->text());
	mwptr->pingResults[mytask]->setCustomWindowsPW(ui->lineEdit_customWindowsPW->text());
	}

// This routine takes the settings from a pingResults task and fills in
// the custom configuration form
void IndividualSettings::slot_setTask(int taskno)
	{
	mytask = taskno;

// Remote connection
	if(mwptr->pingResults[mytask]->getRDPstatus())
		ui->radioButton_useRDP->setChecked(true);
	else
		ui->radioButton_useVNC->setChecked(true);
	ui->comboBox_os->setCurrentIndex(mwptr->pingResults[mytask]->getCustomOS() + 1);
	ui->checkBox_slowPC->setChecked(mwptr->pingResults[mytask]->getSlowness());

// VNC credentials
	switch(mwptr->pingResults[mytask]->getVNCpwType())
		{
		case VNC_DEFAULT:
			ui->radioButton_VNCdefaultCreds->setChecked(true);
			break;
		case VNC_COMMON:
			ui->radioButton_VNCpw->setChecked(true);
			break;
		case VNC_MS_DOMAIN:
			ui->radioButton_VNCmsLogin->setChecked(true);
			break;
		case VNC_CUSTOM:
			ui->radioButton_customVNCpassword->setChecked(true);
		}
	ui->lineEdit_customVNCpassword->setText(mwptr->pingResults[mytask]->getCustomVNCpassword());

// Login warning present
	switch(mwptr->pingResults[mytask]->getWarnSetting())
		{
		case WARN_DEFAULT:
			ui->radioButton_warningDefault->setChecked(true);
			break;
		case WARN_YES:
			ui->radioButton_warningYes->setChecked(true);
			break;
		case WARN_NO:
			ui->radioButton_warningNo->setChecked(true);
			break;
		}

// Remote desktop options
	ui->lineEdit_heightRDP->setText(QString::number(mwptr->pingResults[mytask]->getRDPsize().y()));
	ui->lineEdit_widthRDP->setText(QString::number(mwptr->pingResults[mytask]->getRDPsize().x()));

// Windows login
	ui->checkBox_useWelcome->setChecked(mwptr->pingResults[mytask]->getWelcomeStatus());
	ui->spinBox_welcomePosition->setValue(mwptr->pingResults[mytask]->getWelcomeUserNumber());
	ui->lineEdit_customWindowsName->setText(mwptr->pingResults[mytask]->getCustomWindowsName());
	ui->lineEdit_customWindowsPW->setText(mwptr->pingResults[mytask]->getCustomWindowsPW());
	}

// Wipe out custom settings for this PC
void IndividualSettings::slot_clearSettingsClicked()
	{
	QString host = mwptr->pingResults[mytask]->getHost();
	if(!host.isEmpty())
		{
		QSettings mvc("WSS-DDC", "MultiVNC-Qt\\perMachine\\" + host);
		mvc.clear();
		slot_closeClicked();
		}
	}

// This routine saves the per-machine settings
void IndividualSettings::slot_saveClicked()
	{
	updatePwOption();
	QString host = mwptr->pingResults[mytask]->getHost();
	if(!host.isEmpty())
		{
		QSettings mvc("WSS-DDC", "MultiVNC-Qt\\perMachine\\" + host);
// Remote connection
		mvc.setValue("useRDP", mwptr->pingResults[mytask]->getRDPstatus());
		if(mwptr->pingResults[mytask]->getCustomOS() < 0)
			mvc.setValue("customOS", OS_DEFAULT);
		else
			mvc.setValue("customOS", mwptr->osList[mwptr->pingResults[mytask]->getCustomOS()]);
		mvc.setValue("slowPC", mwptr->pingResults[mytask]->getSlowness());

// VNC credentials
		mvc.setValue("VNCpwType", mwptr->pingResults[mytask]->getVNCpwType());
		mvc.setValue("CustomVNCpassword", encryptString(mwptr->pingResults[mytask]->getCustomVNCpassword()));

// Login warning present
		if(ui->radioButton_warningDefault->isChecked())
			mvc.setValue("warnSetting", WARN_DEFAULT);
		else if(ui->radioButton_warningYes->isChecked())
			mvc.setValue("warnSetting", WARN_YES);
		else
			mvc.setValue("warnSetting", WARN_NO);

// Remote desktop options
		mvc.setValue("RDPsize", mwptr->pingResults[mytask]->getRDPsize());

// Windows login
		mvc.setValue("useWelcome", mwptr->pingResults[mytask]->getWelcomeStatus());
		mvc.setValue("welcomeUserNumber", mwptr->pingResults[mytask]->getWelcomeUserNumber());
		mvc.setValue("CustomWindowsName", mwptr->pingResults[mytask]->getCustomWindowsName());
		mvc.setValue("CustomWindowsPW", encryptString(mwptr->pingResults[mytask]->getCustomWindowsPW()));
		}
	close();
	}
