// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#include "mainwindow.h"
#include "pingresult.h"
#include "ui_pingresult.h"

//#include <comdef.h>

#include <QThread>
#include <QtDebug>

BOOL CALLBACK EnumWindowsProc(HWND hWnd, LPARAM lParam);

HWND PingResult::findWindow(QString title, bool ipOnly)
	{
	foundWindow = nullptr;
	wantedTitle = title;
	if(ipOnly)
		{
		ipInVNCtitle = true;
		wantedTitle = "( " + wantedTitle + " )";
		}
	else
		{
// UltraVNC shows the netbios name which is limited to 15 characters.
		wantedTitle.truncate(15);
// Add a space so we don't get an unwanted match when one full name matches the start of another
		wantedTitle += " ";
		}

	EnumWindows(EnumWindowsProc, 0);
	return foundWindow;
	}

BOOL CALLBACK EnumWindowsProc(HWND hWnd, LPARAM /* lParam */)
	{
	int length = ::GetWindowTextLength(hWnd);
	if(0 == length)
		return TRUE;

	TCHAR *buffer;
	buffer = new TCHAR[static_cast<unsigned int>(length) + 1];
	memset(buffer, 0, static_cast<unsigned int>(length + 1) * sizeof( TCHAR ) );

	GetWindowText( hWnd, buffer, length + 1 );
	QString title =  QString::fromStdWString(buffer);
	delete[] buffer;
	if(ipInVNCtitle)
		{
		if(title.contains(wantedTitle))
			{
			foundWindow = hWnd;
			return false;
			}
		}
	else if(title.startsWith(wantedTitle))
		{
		foundWindow = hWnd;
		return false;
		}

	return true;
	}

// To reduce probability of initial characters being lost, or a garbage character
// getting added at the beginning, send 3 spaces and 4 backspaces before the string
void PingResult::sendQstring(QString str)
	{
	short spacekey = VkKeyScanEx(' ', nullptr);
	short bskey = VkKeyScanEx('\b', nullptr);
	for(int i=0; i<3; i++)
		sendKey(spacekey, false, spacekey);
	for(int i=0; i<4; i++)
		sendKey(bskey, false, bskey);
	for(int i=0; i<str.length(); i++)
		{
		WCHAR tc = static_cast<WCHAR>(str[i].toLatin1());
		short virtkey = VkKeyScanEx(tc, nullptr);
		sendKey(virtkey, false, virtkey);
		}
	}

//	https://forums.codeguru.com/showthread.php?377393-Windows-SDK-User-Interface-How-can-I-emulate-keyboard-events-in-an-application

// Shift state is in high byte of shiftState, low byte may or may not be empty.
// Pace the sending of keys by pausing KEYSTROKE_DELAY_MS milliseconds before each send.
// Sending keys into remote desktop only works with scan codes, pretty much everything
// else would work with virtual key codes.
void PingResult::sendKey(short vk, bool bExtended, short shiftState)
	{
	KEYBDINPUT kb={0, 0, 0, 0, 0};
	INPUT Input={0, {0, 0, 0, 0, 0, 0}};

// If flagged as slow PC, double wait time
	int delayMult = getSlowness() ? 2 : 1;

	vk &= 0xff;
// generate down
	Input.type = INPUT_KEYBOARD;

// Shift if necessary
	if(shiftState & VK_SHIFT_PRESSED)
		{
		kb.wVk = VK_SHIFT;
		kb.dwFlags = KEYEVENTF_SCANCODE;
		kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
//		qDebug() << "Scancode for shift =" << kb.wScan;
		Input.ki = kb;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//		qDebug() << "Shift on";
		SendInput(1, &Input, sizeof(Input));
		}

// Control if necessary
	if(shiftState & VK_CONTROL_PRESSED)
		{
		kb.wVk = VK_CONTROL;
		kb.dwFlags = KEYEVENTF_SCANCODE;
		kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
//		qDebug() << "Scancode for control =" << kb.wScan;
		Input.ki = kb;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//		qDebug() << "Control on";
		SendInput(1, &Input, sizeof(Input));
		}

// Alt if necessary
	if(shiftState & VK_ALT_PRESSED)
		{
		kb.wVk = VK_MENU;
		kb.dwFlags = KEYEVENTF_SCANCODE;
		kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
//		qDebug() << "Scancode for alt =" << kb.wScan;
		Input.ki = kb;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//		qDebug() << "Alt on";
		SendInput(1, &Input, sizeof(Input));
		}

// Key press
	if(bExtended)
		kb.dwFlags = KEYEVENTF_EXTENDEDKEY;
	kb.dwFlags |= KEYEVENTF_SCANCODE;
	kb.wVk = static_cast<WORD>(vk);
	kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
//	qDebug() << "Scancode for " << vk << " =" << kb.wScan;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//	qDebug() << "Key = " << vk;
	SendInput(1, &Input, sizeof(Input));

	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));

// Key release
	kb.dwFlags = 0;
	if(bExtended)
		kb.dwFlags = KEYEVENTF_EXTENDEDKEY;
	kb.dwFlags |= KEYEVENTF_KEYUP | KEYEVENTF_SCANCODE;
	kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//	qDebug() << "key up";
	SendInput(1, &Input, sizeof(Input));

// Release alt if necessary
	kb.dwFlags = 0;
	if(shiftState & VK_ALT_PRESSED)
		{
		kb.wVk = VK_MENU;
		kb.dwFlags |= KEYEVENTF_KEYUP | KEYEVENTF_SCANCODE;
		kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
		Input.ki = kb;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//		qDebug() << "Alt off";
		SendInput(1, &Input, sizeof(Input));
		}
// Release control if necessary
	if(shiftState & VK_CONTROL_PRESSED)
		{
		kb.wVk = VK_CONTROL;
		kb.dwFlags |= KEYEVENTF_KEYUP | KEYEVENTF_SCANCODE;
		kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
		Input.ki = kb;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//		qDebug() << "Control off";
		SendInput(1, &Input, sizeof(Input));
		}
// Release shift if necessary
	if(shiftState & VK_SHIFT_PRESSED)
		{
		kb.wVk = VK_SHIFT;
		kb.dwFlags |= KEYEVENTF_KEYUP | KEYEVENTF_SCANCODE;
		kb.wScan = static_cast<WORD>(MapVirtualKey(kb.wVk, MAPVK_VK_TO_VSC));
		Input.ki = kb;
		QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
//		qDebug() << "Shift off";
		SendInput(1, &Input, sizeof(Input));
		}
	}

// sleepAfter is milliseconds to sleep after sending ctrl-alt-del
void PingResult::sendCAD(int sleepAfter)
	{
	KEYBDINPUT kb={0, 0, 0, 0, 0};
	INPUT Input={0, {0, 0, 0, 0, 0, 0}};

// If flagged as slow PC, double wait time
	int delayMult = getSlowness() ? 2 : 1;

// generate down
	Input.type = INPUT_KEYBOARD;

	kb.wVk = VK_CONTROL;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
	SendInput(1, &Input, sizeof(Input));
	kb.wVk = VK_MENU;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
	SendInput(1, &Input, sizeof(Input));
	kb.wVk = VK_F4;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
	SendInput(1, &Input, sizeof(Input));

// generate up
	kb.dwFlags =  KEYEVENTF_KEYUP;
	kb.wVk = VK_F4;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
	SendInput(1, &Input, sizeof(Input));
	kb.wVk = VK_MENU;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
	SendInput(1, &Input, sizeof(Input));
	kb.wVk = VK_CONTROL;
	Input.ki = kb;
	QThread::msleep(static_cast<unsigned long>(delayMult*mwptr->keystrokeDelay));
	SendInput(1, &Input, sizeof(Input));

	QThread::msleep(static_cast<unsigned long>(delayMult*sleepAfter));
	}

#if defined(__MINGW32__)
namespace _com_util
	{
	BSTR __stdcall ConvertStringToBSTR(const char *const ascii)
		{
		const unsigned int stringLength = static_cast<unsigned int>(lstrlenA(ascii));
		BSTR bstr = SysAllocStringLen(nullptr, stringLength);

		mbstowcs(bstr, ascii, stringLength + 1);

		return bstr;
		}
	}
#endif //defined(__MINGW32__)
