// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

//This is derived from example code at
// https://msdn.microsoft.com/en-us/library/aa390422%28v=vs.85%29.aspx
//#define _WIN32_DCOM
//#undef UNICODE
//#define UNICODE

#include "mainwindow.h"
//#include <OAIdl.h>
//#include <winnetwk.h>
#include <strsafe.h>
#include <WbemIdl.h>
#include <wincred.h>

#include <QtDebug>
#include <QHostInfo>
#include <QString>

// returns 0 if successful, 1 if failed for any reason
int MainWindow::isAnyoneLoggedIn(QString host, QString uname, QString pw, QString *result)
	{
	int rc = isAnyoneLoggedInByRDP(host, uname, pw, result);
	if((0 == rc) && ((*result).length() > 0))
		return 0;
//qDebug() << "Falling back to WMI";
//qDebug() << *result;
// WMI doesn't see RDP users, that's why we only use it if above check fails
// (or if check works but doesn't see anyone)
	rc = wmicQuery(bstr_t("Win32_ComputerSystem"), L"UserName", host, uname, pw, result);
	if("Not found" == *result)
		*result = QString("<font color=\"green\">No one is logged on</font>");
	return rc;
	}

// returns 0 if successful, 1 if failed for any reason
int MainWindow::upTime(QString host, QString uname, QString pw, QString *result)
	{
	int rc = wmicQuery(bstr_t("CIM_OperatingSystem"), L"LastBootUpTime", host, uname, pw, result);
	if("Not found" == *result)
		*result = QString("Boot time not found");
	return rc;
	}

// returns 0 if successful, 1 if failed for any reason
int MainWindow::wmicQuery(bstr_t alias, LPCWSTR property, QString host, QString uname, QString pw, QString *result)
	{
	if(0 == QString::compare(host, QHostInfo::localHostName(), Qt::CaseInsensitive))
		{
		*result = QString("localhost, not tested");
		return 1;
		}
	QString userWithDomain;
	QString userWithoutDomain;
	QString userDomain;
	if(uname.contains("\\"))
		{
		QStringList userParts = uname.split("\\");
		userDomain = userParts[0];
		userWithoutDomain = userParts[1];
		userWithDomain = uname;
		}
	else
		{
		userDomain = host;
		userWithoutDomain = uname;
		userWithDomain = host +"\\" + uname;
		}
qDebug() << "User =" << userWithoutDomain << userWithDomain << ", Password =" << pw;
	std::wstring cimv2 = (QString("\\\\") + host + QString("\\root\\cimv2")).toStdWString();
	std::wstring userWithDomainStd = userWithDomain.toStdWString();
	std::wstring userWithoutDomainStd = userWithoutDomain.toStdWString();
	std::wstring userDomainStd = userDomain.toStdWString();
	std::wstring pwStd = pw.toStdWString();
	bool CoCloseNeeded = false;
	[[maybe_unused]]HRESULT hres;

// Step 1: --------------------------------------------------
// Initialize COM. ------------------------------------------
// This will usually fail because it has already been done
	CoCloseNeeded = !FAILED(CoInitializeEx(nullptr, COINIT_MULTITHREADED));
//	if(!CoCloseNeeded)
//		qDebug() << "CoInitializeEx failed as expected";

// Step 2: --------------------------------------------------
// Set general COM security levels --------------------------
// This is only needed once per process and failure is probably ok
	if(!CoInitializeSecurityCalled)
		{
		hres =  CoInitializeSecurity(
			nullptr,
			-1,                          // COM authentication
			nullptr,                     // Authentication services
			nullptr,                     // Reserved
			RPC_C_AUTHN_LEVEL_DEFAULT,   // Default authentication
			RPC_C_IMP_LEVEL_IDENTIFY,    // Default Impersonation
			nullptr,                     // Authentication info
			EOAC_NONE,                   // Additional capabilities
			nullptr                      // Reserved
			);

		CoInitializeSecurityCalled = true;
		}

// Step 3: ---------------------------------------------------
// Obtain the initial locator to WMI -------------------------

	IWbemLocator *pLoc = nullptr;

	hres = CoCreateInstance(
		CLSID_WbemLocator,
		nullptr,
		CLSCTX_INPROC_SERVER,
		IID_IWbemLocator, reinterpret_cast<LPVOID *>(&pLoc)
		);

	if (FAILED(hres))
		{
		if(CoCloseNeeded)
			CoUninitialize();
		*result = QString("Failed to create IWbemLocator object.");
		return 1;                 // Program has failed.
		}

// Step 4: -----------------------------------------------------
// Connect to WMI through the IWbemLocator::ConnectServer method

	IWbemServices *pSvc = nullptr;

	wchar_t pszPwd[CREDUI_MAX_PASSWORD_LENGTH+1] = {0};
	wchar_t pszDomain[CREDUI_MAX_USERNAME_LENGTH+1];
	wchar_t pszUserName[CREDUI_MAX_USERNAME_LENGTH+1];

// Connect to the remote root\cimv2 namespace
// and obtain pointer pSvc to make IWbemServices calls.
//---------------------------------------------------------

// This call can hang.  WBEM_FLAG_CONNECT_USE_MAX_WAIT limits the hang
// to 2 minutes or less.
	hres = pLoc->ConnectServer(
		_bstr_t(cimv2.c_str()),
		_bstr_t(userWithDomainStd.c_str()),		// User name
		_bstr_t(pwStd.c_str()),				// User password
		nullptr,									// Locale
		WBEM_FLAG_CONNECT_USE_MAX_WAIT,	// Security flags
		nullptr,									// Authority
		nullptr,									// Context object
		&pSvc										// IWbemServices proxy
		);

	if (FAILED(hres))
		{
//		cout << "Could not connect. Error code = 0x"
//			  << hex << hres << endl;
		pLoc->Release();
		if(CoCloseNeeded)
			CoUninitialize();
		*result = QString("Could not connect");
		return 1;                // Program has failed.
		}

//	qDebug() << "Connected to ROOT\\CIMV2 WMI namespace";

// step 5: --------------------------------------------------
// Create COAUTHIDENTITY that can be used for setting security on proxy

	COAUTHIDENTITY *userAcct =  nullptr ;
	COAUTHIDENTITY authIdent;

	 memset(&authIdent, 0, sizeof(COAUTHIDENTITY));
	 StringCchCopy(pszPwd, CREDUI_MAX_PASSWORD_LENGTH+1, reinterpret_cast<STRSAFE_LPCWSTR>(pwStd.c_str()));
	 authIdent.PasswordLength = wcslen (pszPwd);
	 authIdent.Password = reinterpret_cast<USHORT*>(pszPwd);

	 StringCchCopy(pszUserName, CREDUI_MAX_USERNAME_LENGTH+1, reinterpret_cast<STRSAFE_LPCWSTR>(userWithoutDomainStd.c_str()));
	 authIdent.User = reinterpret_cast<USHORT*>(pszUserName);
	 authIdent.UserLength = wcslen(pszUserName);

	 StringCchCopy(pszDomain, CREDUI_MAX_USERNAME_LENGTH+1, reinterpret_cast<STRSAFE_LPCWSTR>(userDomainStd.c_str()));
	 authIdent.Domain = reinterpret_cast<USHORT*>(pszDomain);
	 authIdent.DomainLength = wcslen(userDomainStd.c_str());
	 authIdent.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE;

	 userAcct = &authIdent;

// Step 6: --------------------------------------------------
// Set security levels on a WMI connection ------------------

	hres = CoSetProxyBlanket(
		pSvc,                           // Indicates the proxy to set
		RPC_C_AUTHN_DEFAULT,            // RPC_C_AUTHN_xxx
		RPC_C_AUTHZ_DEFAULT,            // RPC_C_AUTHZ_xxx
		COLE_DEFAULT_PRINCIPAL,         // Server principal name
		RPC_C_AUTHN_LEVEL_PKT_PRIVACY,  // RPC_C_AUTHN_LEVEL_xxx
		RPC_C_IMP_LEVEL_IMPERSONATE,    // RPC_C_IMP_LEVEL_xxx
		userAcct,		                 // client identity
		EOAC_NONE                       // proxy capabilities
		);

	if (FAILED(hres))
		{
		pSvc->Release();
		pLoc->Release();
		if(CoCloseNeeded)
			CoUninitialize();
		*result = QString("Could not set proxy blanket");
		return 1;               // Program has failed.
		}

// Step 7: --------------------------------------------------
// Use the IWbemServices pointer to make requests of WMI ----

	IEnumWbemClassObject* pEnumerator = nullptr;
	hres = pSvc->ExecQuery(
		bstr_t("WQL"),
		bstr_t("Select * from ") + alias,
		WBEM_FLAG_FORWARD_ONLY | WBEM_FLAG_RETURN_IMMEDIATELY,
		nullptr,
		&pEnumerator);

	if (FAILED(hres))
		{
		pSvc->Release();
		pLoc->Release();
		if(CoCloseNeeded)
			CoUninitialize();
		*result = bstr_t("Query for ") + alias + bstr_t(" failed");
		return 1;               // Program has failed.
		}

// Step 8: -------------------------------------------------
// Secure the enumerator proxy

	hres = CoSetProxyBlanket(pEnumerator,
								  RPC_C_AUTHN_DEFAULT,
								  RPC_C_AUTHN_DEFAULT,
								  COLE_DEFAULT_PRINCIPAL,
								  RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
								  RPC_C_IMP_LEVEL_IMPERSONATE,
								  userAcct,
								  EOAC_NONE);

	if (FAILED(hres))
		 {
		 pEnumerator->Release();
		 pSvc->Release();
		 pLoc->Release();
		 if(CoCloseNeeded)
			 CoUninitialize();
		 QString hexresult;
		 hexresult.setNum(hres,16);
		 *result = QString("Could not set proxy blanket on enumerator, error code = 0x%1").arg(hexresult);
		 return 1;               // Program has failed.
		 }

// Step 9: -------------------------------------------------
// Get the data from the query in step 7 -------------------

	 IWbemClassObject *pclsObj = nullptr;
	 ULONG uReturn = 0;

	 while (pEnumerator)
		 {
		 [[maybe_unused]]HRESULT hr = pEnumerator->Next(WBEM_INFINITE, 1,
												  &pclsObj, &uReturn);

		 if(0 == uReturn)
			 break;

		 VARIANT vtProp;

// Get the value of the requested property
		 hr = pclsObj->Get(property, 0, &vtProp, nullptr, nullptr);
		 if(!(SUCCEEDED(hr) && (V_VT(&vtProp) == VT_BSTR)))
			 *result = QString("Not found");
		 else
			 *result = QString(_com_util::ConvertBSTRToString(vtProp.bstrVal));

		 VariantClear(&vtProp);
		 pclsObj->Release();
		 pclsObj = nullptr;
		 break;
		 }

// Cleanup
// ========

	pSvc->Release();
	pLoc->Release();
	if (pEnumerator ) { pEnumerator->Release(); }
	if( pclsObj ) { pclsObj->Release(); }

	if(CoCloseNeeded)
		CoUninitialize();
	return 0;   // Program successfully completed.
	}

#if defined(__MINGW32__)
	namespace _com_util
		{
#endif //defined(__MINGW32__)
char * __stdcall ConvertBSTRToString(BSTR bstr)
	{
	const unsigned int stringLength = static_cast<unsigned int>(lstrlenW(bstr));
	char *const ascii = new char [stringLength + 1];

	size_t ignored;
	wcstombs_s(&ignored, ascii, stringLength + 1, bstr, stringLength + 1);

	return ascii;
	}
#if defined(__MINGW32__)
		}
#endif //defined(__MINGW32__)
